function [xk,p,enoise,delta,info,Yk,res,muv,Ymu] = cosegk(A,b,opts)
%COSEGK	COSE regularization method via Golub-Kahan bidiagonalization
%   y=cosegk(A,b) computes the regularized solution of the least
%   squares problem min||x||, subject to min||A*x-b||.
%
%   To solve the problem min||x-x0||, subject to min||A*x-b||, the user
%   must set opts.x0 to the actual value of x0 and call y=cosegk(A,b,opts).
%
%   [y k enoise delta info Y rho lambda Z]=cosegk(...) returns also the LSQR
%   regularization parameter k, the estimated noise level enoise, the
%   quantities delta being minimized, info on convergence, the LSQR solutions
%   Y, the corresponding residuals rho, the Tikhonov parameters lambda and
%   solutions Z.
%
%   [...]=cosegk(A,b,opts) allows to specify other informations:
%   opts.absrel='abs','rel' switches between minimization of absolute or
%   relative errors (def. 'abs'), opts.show sets the number of the figure where
%   to plot intermediate results (def. no plot), opts.nmax is the max number of
%   iterations (def. 50), opts.tau the convergence tolerance (def. 1e-4).

%   M. E. Hochstenbach, L. Reichel, and G. Rodriguez. 
%   Regularization parameter determination for discrete ill-posed problems. 
%   J. Comput. Appl. Math., 273:132-149, 2015. 
%   https://doi.org/10.1016/j.cam.2014.06.004
%
%   Y. Park, L. Reichel, G. Rodriguez, and X. Yu. 
%   Parameter determination for Tikhonov regularization problems in general
%   form. J. Comput. Appl. Math., 2018. 
%   https://doi.org/10.1016/j.cam.2018.04.049

if nargin < 3 || isempty(opts), opts = struct('empty','empty'); end
if nargin < 2, error('Too few input arguments.'); end

nam = fieldnames(opts);
% set the x0 vector
if ~any(strcmpi('x0',nam)), opts.x0 = []; end
% minimization of absolute or relative errors
if ~any(strcmpi('absrel',nam)), opts.absrel = 'absolute'; end
% show intermediate results
if ~any(strcmpi('show',nam)), opts.show = 0; end
% convergence tolerance
if ~any(strcmpi('tau',nam)), opts.tau = 1e-4; end
% max number of iterations
if ~any(strcmpi('nmax',nam)), opts.nmax = 50; end

tau = opts.tau; nmax = opts.nmax;

[m,n] = size(A);
m1 = size(b,1);
if m ~= m1
	error('Size of A and b must agree.')
end

if ~isempty(opts.x0)
	if ~all(size(opts.x0)==[n,1])
		error('X0 option must be a column vector of lenght n.')
	end
	b = b-A*opts.x0;
end

if isempty(opts.absrel) || (lower(opts.absrel(1))=='a')
	relative = 0;
elseif (lower(opts.absrel(1))=='r')
	relative = 1;
else
	error('ABSREL must be set to either ''relative'' or ''absolute''.')
end

if nargout > 5,  keepsols = 1;  else  keepsols = 0;  end

m = size(A,1);
n = size(A,2);
mvprod = 0;
nrmb = norm(b);
nmax = min(nmax,n);
info = 0;
ncount = 4;

U = zeros(m,nmax+1);
V = zeros(n,nmax);
rho = zeros(nmax,1);
sigma = zeros(nmax+1,1);
res = zeros(nmax,1);
muv = zeros(nmax,1);
delta = zeros(nmax,1);
if keepsols
	Yk = zeros(nmax,nmax);
	Ymu = zeros(nmax,nmax);
end

% perform one step of GK bidiagonalization
sigma(1) = nrmb;
U(:,1) = b/sigma(1);
Vtilde = A'*U(:,1);
mvprod = mvprod+1;
rho(1) = norm(Vtilde);
V(:,1) = Vtilde/rho(1);
Utilde = A*V(:,1)-rho(1)*U(:,1);
mvprod = mvprod+1;
sigma(2) = norm(Utilde);
U(:,2) = Utilde/sigma(2);
C_bar = [rho(1);sigma(2)];

% compute LSQR solution and residual
yk = C_bar \ [nrmb;0];
res(1) = norm(C_bar*yk - [nrmb;0]);
if keepsols,  Yk(1,1) = yk;  end

% compute Tikhonov solution and delta
l = 1;
mu = 1;
ymu = [C_bar;mu] \ [nrmb;0;0];
goon = 1;
while goon && (l < nmax)
	l = l+1;
	yold = ymu;
	[U,V,rho,sigma,mvprod] = onemorestep(A,U,V,rho,sigma,l,mvprod);
	C_bar = [diag(rho(1:l))+diag(sigma(2:l),-1);zeros(1,l-1) sigma(l+1)];
	ymu = [C_bar;mu*eye(l)] \ [nrmb;zeros(2*l,1)];
	goon = norm([yold;0]-ymu) >= tau*norm(ymu);
end
if l>=nmax,  info = 1;  end
muv(1) = 1 / sqrt(rootfinder(C_bar,nrmb,res(1)));
ymu = [C_bar;muv(1)*eye(l)] \ [nrmb;zeros(2*l,1)];
delta(1) = norm([yk;zeros(l-1,1)]-ymu);
if relative,  delta(1) = delta(1)/norm(yk);  end
if keepsols,  Ymu(1:l,1) = ymu;  end

% start iteration
k = 1;
count = 0;
while (count < ncount) && (k < nmax)
	k = k+1;
	% perform one GK step, if required
	if k > l
		l = l+1;
		[U,V,rho,sigma,mvprod] = onemorestep(A,U,V,rho,sigma,l,mvprod);
	end

	% compute LSQR solution
	C_bar = [diag(rho(1:k))+diag(sigma(2:k),-1);zeros(1,k-1) sigma(k+1)];
	[Q R] = qr(C_bar);
	rhs = Q'*[nrmb;zeros(k,1)];
	yk = R(1:k,1:k) \ rhs(1:k);
	res(k) = abs(rhs(k+1));
	if keepsols,  Yk(1:k,k) = yk;  end

	% compute Tikhonov solution and delta
	mu = muv(k-1);
	C_bar = [diag(rho(1:l))+diag(sigma(2:l),-1);zeros(1,l-1) sigma(l+1)];
	yold = [C_bar(1:l,1:l-1);mu*eye(l-1)] \ [nrmb;zeros(2*(l-1),1)];
	ymu = [C_bar;mu*eye(l)] \ [nrmb;zeros(2*l,1)];
	goon = norm([yold;0]-ymu) >= tau;
	while goon && (l < k+nmax)
		l = l+1;
		yold = ymu;
		[U,V,rho,sigma,mvprod] = onemorestep(A,U,V,rho,sigma,l,mvprod);
		C_bar = [diag(rho(1:l))+diag(sigma(2:l),-1); ...
			zeros(1,l-1) sigma(l+1)];
		ymu = [C_bar;mu*eye(l)] \ [nrmb;zeros(2*l,1)];
		goon = norm([yold;0]-ymu) >= tau*norm(ymu);
	end
	if l>=k+nmax,  info = 1;  end
	muv(k) = 1 / sqrt(rootfinder(C_bar,nrmb,res(k)));
	ymu = [C_bar;muv(k)*eye(l)] \ [nrmb;zeros(2*l,1)];
	delta(k) = norm([yk;zeros(l-k,1)]-ymu);
	if relative,  delta(k) = delta(k)/norm(yk);  end
	if keepsols,  Ymu(1:l,k) = ymu;  end

	% stop condition
	if delta(k) > delta(k-1)
		count = count+1;
	else
		count = 0;
	end
end
if k>=nmax,  info = 1;  end
info = [info mvprod l];

U = U(:,1:l+1);
V = V(:,1:l);
rho = rho(1:l);
sigma = sigma(1:l+1);
res = res(1:k);
muv = muv(1:k);
delta = delta(1:k);
if keepsols
	Yk = Yk(1:k,1:k);
	Ymu = Ymu(1:l,1:k);
end

[md p] = min(delta);
enoise = res(p)/nrmb;
C_bar = [diag(rho(1:p))+diag(sigma(2:p),-1);zeros(1,p-1) sigma(p+1)];
yk = C_bar \ [nrmb;zeros(p,1)];
xk = V(:,1:p)*yk;
%mu = muv(p);
%C_bar = [diag(rho(1:l))+diag(sigma(2:l),-1);zeros(1,l-1) sigma(l+1)];
%ymu = [C_bar;muv(p)*eye(l)] \ [nrmb;zeros(2*l,1)];
%xmu = V(:,1:l)*ymu;

if ~isempty(opts.x0)
	xk = xk + opts.x0;
end

if opts.show
	figure(opts.show)
	hh = semilogy([1:size(delta,1)]',delta,p,delta(p),'.r');
	set(hh(2),'markersize',20)
	title(sprintf('delta(k) - minimum at %d',p))
end


function [U,V,rho,sigma,mvprod] = onemorestep(A,U,V,rho,sigma,l,mvprod);
%Utilde = A*V(:,l-1)-rho(l-1)*U(:,l-1);
%sigma(l) = norm(Utilde);
%U(:,l) = Utilde/sigma(l);
% reorthogonalization of columns of U
for k = 1:l-1
	alpha = U(:,k)'*U(:,l);
	U(:,l) = U(:,l)-alpha*U(:,k);
	normu = norm(U(:,l));
	U(:,l) = U(:,l)/normu;
end
Vtilde = A'*U(:,l)-sigma(l)*V(:,l-1);
mvprod = mvprod+1;
rho(l) = norm(Vtilde);
V(:,l) = Vtilde/rho(l);
% reorthogonalization of columns of V
for k = 1:l-1
	alpha = V(:,k)'*V(:,l);
	V(:,l) = V(:,l)-alpha*V(:,k);
	normu = norm(V(:,l));
	V(:,l) = V(:,l)/normu;
end
Utilde = A*V(:,l)-rho(l)*U(:,l);
mvprod = mvprod+1;
sigma(l+1) = norm(Utilde);
U(:,l+1) = Utilde/sigma(l+1);
if (abs(rho(l))<eps) || (abs(sigma(l+1))<eps)
	warning('breakdown in GK')
end


function beta = rootfinder(Cbar,nrmb,res)
l = size(Cbar,2);
% 1/beta is the regularization parameter
nrmbsq = nrmb^2; ressq = res^2;
% function value phibar_l(beta) is denoted yl, the derivative is ydl
% initial beta = 0
beta = 0;
yl = nrmbsq;
ydl = -2*nrmbsq*Cbar(1,1)^2;
% newton's method
newtcount = 0;
stpbd = 1.0001*ressq;
while yl>stpbd
	fval = yl-ressq;
	fder = ydl;
	betadiff = -fval/fder;
	beta = beta+betadiff;
	%update function value and derivative
	elp1 = zeros(2*l+1,1);
	elp1(l+1,1) = 1.0;
	ytmp = [sqrt(beta)*Cbar';eye(l+1)]\elp1;
	yl = nrmbsq*ytmp'*ytmp;
	elp2 = [zeros(l,1);ytmp];
	ztmp = [sqrt(beta)*Cbar';eye(l+1)]\elp2;
	ydl = -2*yl/beta+ 2*nrmbsq*ztmp'*ytmp/beta;
	newtcount = newtcount+1;
end 

