function [yk,ki,enoise,delta,Y,res,lambda,Z,mvp] = cosegkgt(A,L,b,opts)
%COSEGKGT COSE regularization method via Golub-Kahan bidiagonalization
%   y=cosegkgt(A,L,b,kmax) computes the regularized solution of the least
%   squares problem min||L*x||, subject to min||A*x-b||, where L is a
%   regularization matrix. For a standard form problem, L may be set to eye(n).
%
%   To solve the problem min||L*(x-x0)||, subject to min||A*x-b||, the user
%   must set opts.x0 to the actual value of x0 and call y=cosegkgt(A,L,b,opts).
%
%   [y,k,enoise,delta,Y,rho,lambda,Z,mvp]=cosegkgt(...) returns also the 
%   regularization parameter k, the estimated noise level enoise, the
%   quantities delta being minimized, the solutions Y, the corresponding
%   residuals rho, the Tikhonov parameters lambda and solutions Z, the number
%   of matrix-vector products performed.
%
%   [...]=cosegkgt(A,L,b,kmax,absrel,show,reorth,bdtol) allows to specify other
%   informations: absrel='abs','rel' switches between minimization of absolute
%   or relative errors (def. 'abs'), show sets the number of the figure where
%   to plot intermediate results (def. no plot), reorth denotes the kind of
%   reorthogonalization (0 no reorth (def.), 1 MGS, 2 CGS, 3 double CGS), bdtol
%   is the tolerance to detect the breakdown of Golub-Kahan iteration.

%   M. E. Hochstenbach, L. Reichel, and G. Rodriguez. 
%   Regularization parameter determination for discrete ill-posed problems. 
%   J. Comput. Appl. Math., 273:132-149, 2015. 
%   https://doi.org/10.1016/j.cam.2014.06.004
%
%   Y. Park, L. Reichel, G. Rodriguez, and X. Yu. 
%   Parameter determination for Tikhonov regularization problems in general
%   form. J. Comput. Appl. Math., 2018. 
%   https://doi.org/10.1016/j.cam.2018.04.049

if nargin < 4 || isempty(opts), opts = struct('empty','empty'); end
if nargin < 3, error('Too few input arguments.'); end

nam = fieldnames(opts);
% set the x0 vector
if ~any(strcmpi('x0',nam)), opts.x0 = []; end
% minimization of absolute or relative errors
if ~any(strcmpi('absrel',nam)), opts.absrel = 'absolute'; end
% show intermediate results
if ~any(strcmpi('show',nam)), opts.show = 0; end
% initial number of GK steps to perform
if ~any(strcmpi('nsteps',nam)), opts.nsteps = 20; end
% max number of GK steps to perform
if ~any(strcmpi('nmax',nam)), opts.nmax = 200; end
% reorthogonalization switch
if ~any(strcmpi('reorth',nam)), opts.reorth = 1; end
% breakdown tolerance
if ~any(strcmpi('bdtol',nam)), opts.bdtol = 1e-12; end

if isempty(opts.absrel) || (lower(opts.absrel(1))=='a')
	relative = 0;
elseif (lower(opts.absrel(1))=='r')
	relative = 1;
else
	error('ABSREL must be set to either ''relative'' or ''absolute''.')
end

[m,n] = size(A);
m1 = size(b,1);
if m ~= m1
	error('Size of A and b must agree.')
end
[pL n1] = size(L);
if n ~= n1
	error('Size of A and L must agree.')
end

if ~isempty(opts.x0)
	if ~all(size(opts.x0)==[n,1])
		error('X0 option must be a column vector of lenght n.')
	end
	b = b-A*opts.x0;
end

mvp = 0;
nrmb = norm(b);
nsteps = min([opts.nsteps,m,n,pL]);
nmax = min([opts.nmax,m,n,pL]);

if nargout > 4, keepy = 1; else, keepy = 0; end
if nargout > 7, keepz = 1; else, keepz = 0; end

V = zeros(n,nsteps);
U = zeros(m,nsteps+1);
C = zeros(nsteps+1,nsteps);
Q = zeros(pL,nsteps);
R = zeros(nsteps);

U(:,1) = b/nrmb;
k = 0;
goon = 1;
while goon
	% perform some steps of GK bidiagonalization
	for l = 1:nsteps
		k = k+1;
		[U,V,C,Q,R,k,mvp,bdwn] = oneGKstep(A,L,U,V,C,Q,R,k,mvp,...
						opts.reorth,opts.bdtol);
		if bdwn, break, end
	end

	V = V(:,1:k);
	U = U(:,1:k+1);
	C = C(1:k+1,1:k);
	rhs = [nrmb;zeros(k,1)];
	Q = Q(:,1:k);
	R = R(1:k,1:k);

	if ~isempty(opts.x0), y0 = opts.x0; opts.x0 = V'*opts.x0; end
	[yk ki enoise delta Y res lambda Z] = cose( C, R, rhs, opts);

	goon = ~bdwn && (ki>0.75*k) && (k<nmax);
	if goon
		V = [V zeros(n,nsteps)];
		U = [U zeros(m,nsteps+1)];
		C(k+nsteps+1,k+nsteps) = 0;
		Q = [Q zeros(pL,nsteps)];
		R(k+nsteps,k+nsteps) = 0;
		warning('cose:bdown','Extending to %d the number of initial iterations',k+nsteps)
	end
end

if ~all(isfinite(lambda)), warning('cose:bdown','Some lambda values are not finite.'), end
yk = V*yk;
if keepy, Y = V*Y; end
if keepz, Z = V*Z; end
if ~isempty(opts.x0)
	yk = yk + y0;
	if keepy, Y = Y + y0*ones(1,size(Y,2)); end
	if keepz, Z = Z + y0*ones(1,size(Z,2)); end
end

function [U,V,C,Q,R,k,mvp,bdwn] = oneGKstep(A,L,U,V,C,Q,R,k,mvp,reorth,bdtol)

sqrteps = sqrt(eps);
bdwn = 0;
vtilde = A'*U(:,k);
mvp = mvp+1;
if k > 1
	vtilde = vtilde - C(k,k-1)*V(:,k-1);
end

% Reorthonormalizing the columns of V
if reorth
	switch reorth
	case 1	% MGS
		for j = 1:k-1
			vtilde = vtilde - (vtilde'*V(:,j))*V(:,j);
		end
	case 2	% CGS
		vtilde = vtilde - V(:,1:k-1)*(V(:,1:k-1)'*vtilde);
	case 3	% double CGS
		c = V(:,1:k-1)'*vtilde;
		vtilde = vtilde - V(:,1:k-1)*c;
		%alfa = alfa + sum(c);		% See irbleigs
		if norm(c) > sqrteps
			c1 = V(:,1:k-1)'*vtilde;
			vtilde = vtilde - V(:,1:k-1)*c1;
			%alfa = alfa + sum(c1);	% See irbleigs
		end
	end
end

C(k,k) = norm(vtilde);

if C(k,k) < bdtol
	warning('cose:bdown','\nRho numerically zero for k=%d\n',k);
	k = k-1;
	bdwn = 1;
	return
else
	V(:,k) = vtilde./C(k,k);
end

utilde = A*V(:,k);
mvp = mvp+1;
utilde = utilde - C(k,k)*U(:,k);

% Reorthonormalizing the columns of U
if reorth
	switch reorth
		case 1	% MGS
			for j = 1:k
				utilde = utilde - (utilde'*U(:,j))*U(:,j);
			end
		case 2	% CGS
			utilde = utilde - U(:,1:k)*(U(:,1:k)'*utilde);
		case 3	% double CGS
			c = U(:,1:k)'*utilde;
			utilde = utilde - U(:,1:k)*c;
			%alfa = alfa + sum(c);		% See irbleigs
			if norm(c) > sqrteps
				c1 = U(:,1:k)'*utilde;
				utilde = utilde - U(:,1:k)*c1;
				%alfa = alfa + sum(c1);	% See irbleigs
			end
	end
end

C(k+1,k) = norm(utilde);

if C(k+1,k) < bdtol
	warning('cose:bdown','\nSigma numerically zero for k=%d\n',k);
	U(:,k+1) = zeros(size(A,1),1);
	bdwn = 1;
else
	U(:,k+1) = utilde./C(k+1,k);
end

%[Q(:,1:k) R(1:k,1:k)] = qr(L*V(:,1:k),0);
[Q R] = updateqr(L*V(:,k),k,Q,R);



function [Q,R] = updateqr(v,n,Q,R)

if any(abs(R(:,n))), error('!!!'), end
if n == 1
	R(1,1) = norm(v);
	Q(:,1) = v/R(1,1);
else
	R(1:n-1,n) = Q(:,1:n-1)'*v;
	v = v-Q(:,1:n-1)*R(1:n-1,n);
	% which is the best?
	%for j = 1:n-1
	%	v = v - (v'*Q(:,j))*Q(:,j);
	%end
	R(n,n) = norm(v);
	Q(:,n) = v/R(n,n);
end

